/***********************************************************************
  This file is part of HA, a general purpose file archiver.
  Copyright (C) 1995 Harri Hirvola

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
************************************************************************
	HA ASC method
***********************************************************************/

#include <proto/exec.h>
#include "ha.h"
#include "haio.h"
#include "asc.h"
#include "swdict.h"
#include "acoder.h"


static void tabinit(U16B t[], U16B tl, U16B ival) {

    register U16B i,j;
    
    for (i=tl;i<2*tl;++i) t[i]=ival;
    for (i=tl-1,j=(tl<<1)-2;i;--i,j-=2) {
	t[i]=t[j]+t[j+1];
    }
	 t[0]=0;	/* bugfix */
}

static void tscale(U16B t[], U16B tl) {

    register U16B i,j;
    
    for (i=(tl<<1)-1;i>=tl;--i) {
	if (t[i]>1) t[i]>>=1;
    }
    for (i=tl-1,j=(tl<<1)-2;i;--i,j-=2) {
	t[i]=t[j]+t[j+1];
    }
}

static void tupd(U16B t[], U16B tl, U16B maxt, U16B step, U16B p) {

    register S16B i;
    
    for (i=p+tl;i;i>>=1) t[i]+=step;
    if (t[1]>=maxt) tscale(t,tl);
}

static void tzero(U16B t[], U16B tl, U16B p) {

    register S16B i,step;
    
    for (i=p+tl,step=t[i];i;i>>=1) t[i]-=step;
}

static void ttscale(struct asc *asc, U16B con) {

    asc->ttab[con][0]>>=1;
    if (asc->ttab[con][0]==0) asc->ttab[con][0]=1;
    asc->ttab[con][1]>>=1;
    if (asc->ttab[con][1]==0) asc->ttab[con][1]=1;
}

static void codepair(struct asc *asc, S16B l, S16B p) {

    register U16B i,j,lt,k,cf,tot;
    
    i=asc->ttab[asc->ttcon][0]+asc->ttab[asc->ttcon][1];
    ac_out(asc->ac,asc->haio,asc->ttab[asc->ttcon][0],i,i+1);
    asc->ttab[asc->ttcon][1]+=TTSTEP;
    if (i>=MAXTT) ttscale(asc,asc->ttcon);
    asc->ttcon=((asc->ttcon<<1)|1)&TTOMASK;
    while (asc->ccnt>asc->pmax) {
	tupd(asc->ptab,PTCODES,MAXPT,PTSTEP,asc->npt++);
	asc->pmax<<=1;	
    }

    for (i=p,j=0;i;++j,i>>=1); 
    cf=asc->ptab[PTCODES+j];
    tot=asc->ptab[1];
    for (lt=0,i=PTCODES+j;i;i>>=1) {
	if (i&1) lt+=asc->ptab[i-1];
	asc->ptab[i]+=PTSTEP;
    }
    if (asc->ptab[1]>=MAXPT) tscale(asc->ptab,PTCODES);
    ac_out(asc->ac,asc->haio,lt,lt+cf,tot);
    if (p>1) {
	for (i=0x8000U;!(p&i);i>>=1);
	j=p&~i;
	if (i!=(asc->pmax>>1)) ac_out(asc->ac,asc->haio,j,j+1,i);
	else ac_out(asc->ac,asc->haio,j,j+1,asc->ccnt-(asc->pmax>>1));
    }
    i=l-MINLEN;
    if (i==LENCODES-1) i=SLCODES-1,j=0xffff;
    else if (i<SLCODES-1) j=0xffff; 
    else {
	j=(i-SLCODES+1)&LLMASK;
	i=((i-SLCODES+1)>>LLBITS)+SLCODES;	
    }
    if ((cf=asc->ltab[LTCODES+i])==0) {
	ac_out(asc->ac,asc->haio,asc->ltab[1],asc->ltab[1]+asc->les,asc->ltab[1]+asc->les);	
	for (lt=0,k=LTCODES+i;k;k>>=1) {
	    if (k&1) lt+=asc->eltab[k-1];
	    asc->ltab[k]+=LTSTEP;
	}
	if (asc->ltab[1]>=MAXLT) tscale(asc->ltab,LTCODES);
	ac_out(asc->ac,asc->haio,lt,lt+asc->eltab[LTCODES+i],asc->eltab[1]);
	tzero(asc->eltab,LTCODES,i);
	if (asc->eltab[1]!=0) asc->les+=LTSTEP;
	else asc->les=0;
	for (k=i<=LPLEN?0:i-LPLEN;
	     k<(i+LPLEN>=LTCODES-1?LTCODES-1:i+LPLEN);++k) {
	    if (asc->eltab[LTCODES+k]) tupd(asc->eltab,LTCODES,MAXLT,1,k);
	}
    }
    else {
	tot=asc->ltab[1]+asc->les;
	for (lt=0,k=LTCODES+i;k;k>>=1) {
	    if (k&1) lt+=asc->ltab[k-1];
	    asc->ltab[k]+=LTSTEP;
	}
	if (asc->ltab[1]>=MAXLT) tscale(asc->ltab,LTCODES);
	ac_out(asc->ac,asc->haio,lt,lt+cf,tot);
    }
    if (asc->ltab[LTCODES+i]==LCUTOFF) asc->les-=LTSTEP<asc->les?LTSTEP:asc->les-1; 
    if (j!=0xffff) ac_out(asc->ac,asc->haio,j,j+1,LLLEN); 
    if (asc->ccnt<POSCODES) {
	asc->ccnt+=l;
	if (asc->ccnt>POSCODES) asc->ccnt=POSCODES;
    }
}

static void codechar(struct asc *asc, S16B c) {

    register U16B i,lt,tot,cf;

    i=asc->ttab[asc->ttcon][0]+asc->ttab[asc->ttcon][1];
    ac_out(asc->ac,asc->haio,0,asc->ttab[asc->ttcon][0],i+1);
    asc->ttab[asc->ttcon][0]+=TTSTEP;
    if (i>=MAXTT) ttscale(asc,asc->ttcon);
    asc->ttcon=(asc->ttcon<<1)&TTOMASK;
    if ((cf=asc->ctab[CTCODES+c])==0) {
	ac_out(asc->ac,asc->haio,asc->ctab[1],asc->ctab[1]+asc->ces,asc->ctab[1]+asc->ces);
	for (lt=0,i=CTCODES+c;i;i>>=1) {
	    if (i&1) lt+=asc->ectab[i-1];
	    asc->ctab[i]+=CTSTEP;
	}
	if (asc->ctab[1]>=MAXCT) tscale(asc->ctab,CTCODES);
	ac_out(asc->ac,asc->haio,lt,lt+asc->ectab[CTCODES+c],asc->ectab[1]);
	tzero(asc->ectab,CTCODES,c);
	if (asc->ectab[1]!=0) asc->ces+=CTSTEP;
	else asc->ces=0;
	for (i=c<=CPLEN?0:c-CPLEN;
	     i<(c+CPLEN>=CTCODES-1?CTCODES-1:c+CPLEN);++i) {
	    if (asc->ectab[CTCODES+i]) tupd(asc->ectab,CTCODES,MAXCT,1,i);
	}
    }
    else {
	tot=asc->ctab[1]+asc->ces;
	for (lt=0,i=CTCODES+c;i;i>>=1) {
	    if (i&1) lt+=asc->ctab[i-1];
	    asc->ctab[i]+=CTSTEP;
	}
	if (asc->ctab[1]>=MAXCT) tscale(asc->ctab,CTCODES);
	ac_out(asc->ac,asc->haio,lt,lt+cf,tot);
    }
    if (asc->ctab[CTCODES+c]==CCUTOFF) asc->ces-=CTSTEP<asc->ces?CTSTEP:asc->ces-1; 
    if (asc->ccnt<POSCODES) ++asc->ccnt;
}

/***********************************************************************
	Reset routine
***********************************************************************/

static void asc_reset(struct asc *asc) {
	
	register S16B i;

	asc->ces=CTSTEP;
	asc->les=LTSTEP;
	asc->ccnt=0;
	asc->ttcon=0;
	asc->npt=asc->pmax=1;
	for (i=0;i<TTORD;++i) asc->ttab[i][0]=asc->ttab[i][1]=TTSTEP;
	tabinit(asc->ltab,LTCODES,0);
	tabinit(asc->eltab,LTCODES,1);
	tabinit(asc->ctab,CTCODES,0);
	tabinit(asc->ectab,CTCODES,1);
	tabinit(asc->ptab,PTCODES,0);	
	tupd(asc->ptab,PTCODES,MAXPT,PTSTEP,0);
}

/***********************************************************************
	Setup routine
***********************************************************************/

struct asc *asc_setup(void) {
	
	struct asc *asc;

	asc = (struct asc *)AllocVec(sizeof(struct asc),0);

	return asc;
}

/***********************************************************************
	Cleanup routine
***********************************************************************/

void asc_cleanup(struct asc *asc) {
    
	 FreeVec(asc);
}

/***********************************************************************
	Encoding
***********************************************************************/

U32B asc_pack(struct asc *asc, unsigned char *ibuf, U32B ilen, unsigned char *obuf, U32B olen) {
    
    S16B oc;
    U16B omlf,obpos;
    U32B plen=0;
    
	asc_reset(asc);
	if (asc->haio = init_io(ibuf,ilen,obuf,olen))
	{
    if (asc->swd = swd_init(asc->haio,LENCODES+MINLEN-1,POSCODES))
	 {
     if (asc->ac = ac_init_encode(asc->haio))
	  {
    for (swd_findbest(asc->swd,asc->haio);asc->swd->swd_char>=0;) {
	if (asc->swd->swd_mlf>MINLEN || (asc->swd->swd_mlf==MINLEN && asc->swd->swd_bpos<MINLENLIM)) {
	    omlf=asc->swd->swd_mlf;
	    obpos=asc->swd->swd_bpos;
	    oc=asc->swd->swd_char;
	    swd_findbest(asc->swd,asc->haio);
	    if (asc->swd->swd_mlf>omlf) codechar(asc,oc);
	    else {
		swd_accept(asc->swd,asc->haio);
		codepair(asc,omlf,obpos);
		swd_findbest(asc->swd,asc->haio);
	    }
	}
	else {
	    asc->swd->swd_mlf=MINLEN-1;
	    codechar(asc,asc->swd->swd_char);
	    swd_findbest(asc->swd,asc->haio);
	}
    }
    ac_out(asc->ac,asc->haio,asc->ttab[asc->ttcon][0]+asc->ttab[asc->ttcon][1],
	   asc->ttab[asc->ttcon][0]+asc->ttab[asc->ttcon][1]+1,asc->ttab[asc->ttcon][0]+asc->ttab[asc->ttcon][1]+1);
      ac_end_encode(asc->ac,asc->haio);
	  }
     swd_cleanup(asc->swd);
	 }
     plen=end_io(asc->haio);
	 }
    return plen;
}

/***********************************************************************
	Decoding
***********************************************************************/

U32B asc_unpack(struct asc *asc, unsigned char *ibuf, U32B ilen, unsigned char *obuf, U32B olen) {
    
    register U16B l,p,tv,i,lt;
    U32B ulen=0;
    
	asc_reset(asc);
 	if (asc->haio = init_io(ibuf,ilen,obuf,olen))
	{
    if (asc->swd = swd_dinit(asc->haio,POSCODES))
	 {
     if (asc->ac = ac_init_decode(asc->haio))
	  {
    for (;;) {
	tv=ac_threshold_val(asc->ac,asc->haio,asc->ttab[asc->ttcon][0]+asc->ttab[asc->ttcon][1]+1);		
	i=asc->ttab[asc->ttcon][0]+asc->ttab[asc->ttcon][1];
	if (asc->ttab[asc->ttcon][0]>tv) {
	    ac_in(asc->ac,asc->haio,0,asc->ttab[asc->ttcon][0],i+1);
	    asc->ttab[asc->ttcon][0]+=TTSTEP;
	    if (i>=MAXTT) ttscale(asc,asc->ttcon);
	    asc->ttcon=(asc->ttcon<<1)&TTOMASK;
	    tv=ac_threshold_val(asc->ac,asc->haio,asc->ctab[1]+asc->ces);
	    if (tv>=asc->ctab[1]) {
		ac_in(asc->ac,asc->haio,asc->ctab[1],asc->ctab[1]+asc->ces,asc->ctab[1]+asc->ces);
		tv=ac_threshold_val(asc->ac,asc->haio,asc->ectab[1]);
		for (l=2,lt=0;;) {
		    if (lt+asc->ectab[l]<=tv) {
			lt+=asc->ectab[l];
			++l;
		    }
		    if (l>=CTCODES) {
			l-=CTCODES;
			break;
		    }
		    l<<=1;
		}
		ac_in(asc->ac,asc->haio,lt,lt+asc->ectab[CTCODES+l],asc->ectab[1]);
		tzero(asc->ectab,CTCODES,l);
		if (asc->ectab[1]!=0) asc->ces+=CTSTEP;
		else asc->ces=0;
		for (i=l<CPLEN?0:l-CPLEN;
		     i<(l+CPLEN>=CTCODES-1?CTCODES-1:l+CPLEN);++i) {
		    if (asc->ectab[CTCODES+i]) tupd(asc->ectab,CTCODES,MAXCT,1,i);
		}
	    }
	    else {
		for (l=2,lt=0;;) {
		    if (lt+asc->ctab[l]<=tv) {
			lt+=asc->ctab[l];
			l++;
		    }
		    if (l>=CTCODES) {
			l-=CTCODES;
			break;
		    }
		    l<<=1;
		}
		ac_in(asc->ac,asc->haio,lt,lt+asc->ctab[CTCODES+l],asc->ctab[1]+asc->ces);
	    }
	    tupd(asc->ctab,CTCODES,MAXCT,CTSTEP,l);
	    if (asc->ctab[CTCODES+l]==CCUTOFF) asc->ces-=CTSTEP<asc->ces?CTSTEP:asc->ces-1; 
	    swd_dchar(asc->swd,asc->haio,l);
	    if (asc->ccnt<POSCODES) ++asc->ccnt;
	}
	else if (i>tv) {
	    ac_in(asc->ac,asc->haio,asc->ttab[asc->ttcon][0],i,i+1);
	    asc->ttab[asc->ttcon][1]+=TTSTEP;
	    if (i>=MAXTT) ttscale(asc,asc->ttcon);
	    asc->ttcon=((asc->ttcon<<1)|1)&TTOMASK;
	    while (asc->ccnt>asc->pmax) {
		tupd(asc->ptab,PTCODES,MAXPT,PTSTEP,asc->npt++);
		asc->pmax<<=1;	
	    }
	    tv=ac_threshold_val(asc->ac,asc->haio,asc->ptab[1]);
	    for (p=2,lt=0;;) {
		if (lt+asc->ptab[p]<=tv) {
		    lt+=asc->ptab[p];
		    p++;
		}
		if (p>=PTCODES) {
		    p-=PTCODES;
		    break;
		}
		p<<=1;
	    }
	    ac_in(asc->ac,asc->haio,lt,lt+asc->ptab[PTCODES+p],asc->ptab[1]);
	    tupd(asc->ptab,PTCODES,MAXPT,PTSTEP,p);
	    if (p>1) {
		for (i=1;p;i<<=1,--p);
		i>>=1;
		if (i==(asc->pmax>>1)) l=asc->ccnt-(asc->pmax>>1);
		else l=i;
		p=ac_threshold_val(asc->ac,asc->haio,l);
		ac_in(asc->ac,asc->haio,p,p+1,l);
		p+=i;
	    }
	    tv=ac_threshold_val(asc->ac,asc->haio,asc->ltab[1]+asc->les);
	    if (tv>=asc->ltab[1]) {
		ac_in(asc->ac,asc->haio,asc->ltab[1],asc->ltab[1]+asc->les,asc->ltab[1]+asc->les);
		tv=ac_threshold_val(asc->ac,asc->haio,asc->eltab[1]);
		for (l=2,lt=0;;) {
		    if (lt+asc->eltab[l]<=tv) {
			lt+=asc->eltab[l];
			++l;
		    }
		    if (l>=LTCODES) {
			l-=LTCODES;
			break;
		    }
		    l<<=1;
		}
		ac_in(asc->ac,asc->haio,lt,lt+asc->eltab[LTCODES+l],asc->eltab[1]);
		tzero(asc->eltab,LTCODES,l);
		if (asc->eltab[1]!=0) asc->les+=LTSTEP;
		else asc->les=0;
		for (i=l<LPLEN?0:l-LPLEN;
		     i<(l+LPLEN>=LTCODES-1?LTCODES-1:l+LPLEN);++i) {
		    if (asc->eltab[LTCODES+i]) tupd(asc->eltab,LTCODES,MAXLT,1,i);
		}
	    }
	    else {
		for (l=2,lt=0;;) {
		    if (lt+asc->ltab[l]<=tv) {
			lt+=asc->ltab[l];
			++l;
		    }
		    if (l>=LTCODES) {
			l-=LTCODES;
			break;
		    }
		    l<<=1;
		}
		ac_in(asc->ac,asc->haio,lt,lt+asc->ltab[LTCODES+l],asc->ltab[1]+asc->les);
	    }
	    tupd(asc->ltab,LTCODES,MAXLT,LTSTEP,l);
	    if (asc->ltab[LTCODES+l]==LCUTOFF) asc->les-=LTSTEP<asc->les?LTSTEP:asc->les-1; 
	    if (l==SLCODES-1) l=LENCODES-1;
	    else if (l>=SLCODES) {
		i=ac_threshold_val(asc->ac,asc->haio,LLLEN);
		ac_in(asc->ac,asc->haio,i,i+1,LLLEN);
		l=((l-SLCODES)<<LLBITS)+i+SLCODES-1;
	    }
	    l+=3;
	    if (asc->ccnt<POSCODES) {
		asc->ccnt+=l;
		if (asc->ccnt>POSCODES) asc->ccnt=POSCODES;
	    }
	    swd_dpair(asc->swd,asc->haio,l,p);				
	}
	else {
       break;
	}
    }
     ac_in(asc->ac,asc->haio,i,i+1,i+1);
       ac_end_decode(asc->ac,asc->haio);
		}
      swd_cleanup(asc->swd);
	  }
     ulen=end_io(asc->haio);
	 }
    return ulen;
}
 
